/*
 * SuperTrak EX8350/8300/16350/16300 Storage Controller driver for Linux
 *
 *	Copyright (C) 2005, 2006 Promise Technology Inc.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	Written By:
 *		Ed Lin <promise_linux@promise.com>
 *
 *	Version: 2.24.0.4
 *
 */

#include <linux/init.h>
#include <stdarg.h>
#include <asm/io.h>
#include <asm/string.h>
#include <asm/irq.h>
#include <asm/byteorder.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/irq.h>
#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif
#include <linux/blk.h>
#include <linux/types.h>
#include <linux/hdreg.h>
#include <linux/notifier.h>        
#include <linux/reboot.h>

#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/smp.h>

#include <asm/uaccess.h>
#include <scsi.h>
#include <sd.h>
#include <hosts.h>
#if defined(CONFIG_DISKDUMP) || defined(CONFIG_DISKDUMP_MODULE)
#include <scsi_dump.h>
#endif
#ifdef __x86_64__
#include <asm/ioctl32.h>
#endif
#define STEX_NAME "stex"
#define ST_DRIVER_DATE "(Aug 7, 2006)"
#define ST_DRIVER_VERSION "2.24.0.4"
#define ST_VER_MAJOR 		2
#define ST_VER_MINOR 	 24	
#define ST_OEM			0
#define ST_BUILD_VER 	4	

#ifndef __iomem
#define __iomem

#define  SAM_STAT_GOOD 0x00
#define  SAM_STAT_CHECK_CONDITION 0x02
#define  SAM_STAT_BUSY 0x08
#ifndef DMA_64BIT_MASK
#define DMA_64BIT_MASK 0xffffffffffffffffULL
#endif
#ifndef DMA_32BIT_MASK
#define DMA_32BIT_MASK 0xffffffffULL
#endif
#endif

#ifndef MSEC_PER_SEC
#define MSEC_PER_SEC 1000L

#if HZ <= MSEC_PER_SEC && !(MSEC_PER_SEC % HZ)
#define msleep(m) schedule_timeout((m + (MSEC_PER_SEC / HZ) - 1) / (MSEC_PER_SEC / HZ))
#elif HZ > MSEC_PER_SEC && !(HZ % MSEC_PER_SEC)
#define msleep(m) schedule_timeout(m * (HZ / MSEC_PER_SEC))
#else
#define msleep(m) schedule_timeout((m * HZ + MSEC_PER_SEC - 1) / MSEC_PER_SEC)
#endif
#endif

enum{
	/* MU register offset */
	IMR0	= 0x10,	/* MU_INBOUND_MESSAGE_REG0 */
	IMR1	= 0x14,	/* MU_INBOUND_MESSAGE_REG1 */
	OMR0	= 0x18,	/* MU_OUTBOUND_MESSAGE_REG0 */
	OMR1	= 0x1c,	/* MU_OUTBOUND_MESSAGE_REG1 */
	IDBL	= 0x20,	/* MU_INBOUND_DOORBELL */
	IIS	= 0x24,	/* MU_INBOUND_INTERRUPT_STATUS */
	IIM	= 0x28,	/* MU_INBOUND_INTERRUPT_MASK */
	ODBL	= 0x2c,	/* MU_OUTBOUND_DOORBELL */
	OIS	= 0x30,	/* MU_OUTBOUND_INTERRUPT_STATUS */
	OIM	= 0x3c,	/* MU_OUTBOUND_INTERRUPT_MASK */

	/* MU register value */
	MU_INBOUND_DOORBELL_HANDSHAKE		= 1,
	MU_INBOUND_DOORBELL_REQHEADCHANGED	= 2,
	MU_INBOUND_DOORBELL_STATUSTAILCHANGED	= 4,
	MU_INBOUND_DOORBELL_HMUSTOPPED		= 8,
	MU_INBOUND_DOORBELL_RESET		= 16,

	MU_OUTBOUND_DOORBELL_HANDSHAKE		= 1,
	MU_OUTBOUND_DOORBELL_REQUESTTAILCHANGED	= 2,
	MU_OUTBOUND_DOORBELL_STATUSHEADCHANGED	= 4,
	MU_OUTBOUND_DOORBELL_BUSCHANGE		= 8,
	MU_OUTBOUND_DOORBELL_HASEVENT		= 16,

	/* MU status code */
	MU_STATE_STARTING			= 1,
	MU_STATE_FMU_READY_FOR_HANDSHAKE	= 2,
	MU_STATE_SEND_HANDSHAKE_FRAME		= 3,
	MU_STATE_STARTED			= 4,
	MU_STATE_RESETTING			= 5,

	MU_MAX_DELAY_TIME			= 240000,
	MU_HANDSHAKE_SIGNATURE			= 0x55aaaa55,
	HMU_PARTNER_TYPE			= 2,

	/* firmware returned values */
	SRB_STATUS_SUCCESS			= 0x01,
	SRB_STATUS_ERROR			= 0x04,
	SRB_STATUS_BUSY				= 0x05,
	SRB_STATUS_INVALID_REQUEST		= 0x06,
	SRB_STATUS_SELECTION_TIMEOUT		= 0x0A,
	SRB_SEE_SENSE 				= 0x80,

	/* task attribute */
	TASK_ATTRIBUTE_SIMPLE			= 0x0,
	TASK_ATTRIBUTE_HEADOFQUEUE		= 0x1,
	TASK_ATTRIBUTE_ORDERED			= 0x2,
	TASK_ATTRIBUTE_ACA			= 0x4,

	/* request count, etc. */
	MU_MAX_REQUEST				= 32,
	TAG_BITMAP_LENGTH			= MU_MAX_REQUEST,

	/* one message wasted, use MU_MAX_REQUEST+1
		to handle MU_MAX_REQUEST messages */
	MU_REQ_COUNT				= (MU_MAX_REQUEST + 1),
	MU_STATUS_COUNT				= (MU_MAX_REQUEST + 1),

	STEX_CDB_LENGTH			= MAX_COMMAND_SIZE,
	REQ_VARIABLE_LEN			= 1024,
	STATUS_VAR_LEN				= 128,
	ST_CAN_QUEUE				= MU_MAX_REQUEST,
	ST_CMD_PER_LUN				= MU_MAX_REQUEST,
	ST_MAX_SG				= 32,

	/* sg flags */
	SG_CF_EOT				= 0x80,	/* end of table */
	SG_CF_64B				= 0x40,	/* 64 bit item */
	SG_CF_HOST				= 0x20,	/* sg in host memory */

	ST_MAX_ADAPTORS	= 16,

	ST_MAX_ARRAY_SUPPORTED			= 16,
	ST_MAX_TARGET_NUM			= (ST_MAX_ARRAY_SUPPORTED+1),
	ST_MAX_LUN_PER_TARGET			= 16,
	st_shasta				= 0,
	st_vsc					= 1,
	PASSTHRU_REQ_TYPE			= 0x00000001,
	PASSTHRU_REQ_NO_WAKEUP			= 0x00000100,
	ST_INTERNAL_TIMEOUT			= 30,

	/* vendor specific commands of Promise */
	ARRAY_CMD				= 0xe0,
	CONTROLLER_CMD				= 0xe1,
	DEBUGGING_CMD				= 0xe2,
	PASSTHRU_CMD				= 0xe3,

	PASSTHRU_GET_ADAPTER			= 0x05,
	PASSTHRU_GET_DRVVER			= 0x10,
	CTLR_POWER_STATE_CHANGE			= 0x0e,
	CTLR_POWER_SAVING			= 0x01,

	PASSTHRU_SIGNATURE			= 0x4e415041,

	INQUIRY_EVPD				= 0x01,
};

typedef __u16 __le16;
typedef __u32 __le32;

struct st_sgitem {
	u8 ctrl;	/* SG_CF_xxx */
	u8 reserved[3];
	__le32 count;
	__le32 addr;
	__le32 addr_hi;
};

struct st_sgtable {
	__le16 sg_count;
	__le16 max_sg_count;
	__le32 sz_in_byte;
	struct st_sgitem table[ST_MAX_SG];
};

struct handshake_frame {
	__le32 rb_phy;		/* request payload queue physical address */
	__le32 rb_phy_hi;
	__le16 req_sz;		/* size of each request payload */
	__le16 req_cnt;		/* count of reqs the buffer can hold */
	__le16 status_sz;	/* size of each status payload */
	__le16 status_cnt;	/* count of status the buffer can hold */
	__le32 hosttime;	/* seconds from Jan 1, 1970 (GMT) */
	__le32 hosttime_hi;
	u8 partner_type;	/* who sends this frame */
	u8 reserved0[7];
	__le32 partner_ver_major;
	__le32 partner_ver_minor;
	__le32 partner_ver_oem;
	__le32 partner_ver_build;
	u32 reserved1[4];
};

struct req_msg {
	__le16 tag;
	u8 lun;
	u8 target;
	u8 task_attr;
	u8 task_manage;
	u8 prd_entry;
	u8 payload_sz;		/* payload size in 4-byte */
	u8 cdb[STEX_CDB_LENGTH];
	u8 variable[REQ_VARIABLE_LEN];
};

struct status_msg {
	__le16 tag;
	u8 lun;
	u8 target;
	u8 srb_status;
	u8 scsi_status;
	u8 reserved;
	u8 payload_sz;		/* payload size in 4-byte */
	u8 variable[STATUS_VAR_LEN];
};

#define MU_REQ_BUFFER_SIZE	(MU_REQ_COUNT * sizeof(struct req_msg))
#define MU_STATUS_BUFFER_SIZE	(MU_STATUS_COUNT * sizeof(struct status_msg))
#define MU_BUFFER_SIZE		(MU_REQ_BUFFER_SIZE + MU_STATUS_BUFFER_SIZE)

struct ver_info {
	u32 major;
	u32 minor;
	u32 oem;
	u32 build;
	u32 reserved[2];
};

struct st_frame {
	u32 base[6];
	u32 rom_addr;

	struct ver_info drv_ver;
	struct ver_info bios_ver;

	u32 bus;
	u32 slot;
	u32 irq_level;
	u32 irq_vec;
	u32 id;
	u32 subid;

	u32 dimm_size;
	u8 dimm_type;
	u8 reserved[3];

	u32 channel;
	u32 reserved1;
};

struct st_drvver {
	u32 major;
	u32 minor;
	u32 oem;
	u32 build;
	u32 signature[2];
	u8 console_id;
	u8 host_no;
	u8 reserved0[2];
	u32 reserved[3];
};

struct st_ccb {
	struct req_msg *req;
	struct scsi_cmnd *cmd;

	void *sense_buffer;
	struct page *page;
	unsigned int sense_bufflen;
	unsigned int page_offset;

	u32 req_type;
	u8 srb_status;
	u8 scsi_status;
};

struct st_hba {
	void __iomem *mmio_base;	/* iomapped PCI memory space */
	void *dma_mem;
	dma_addr_t dma_handle;

	struct Scsi_Host *host;
	struct pci_dev *pdev;

	u32 tag;
	u32 req_head;
	u32 req_tail;
	u32 status_head;
	u32 status_tail;

	struct status_msg *status_buffer;
	struct st_ccb ccb[MU_MAX_REQUEST];
	struct st_ccb *wait_ccb;
	wait_queue_head_t waitq;

	struct timer_list ioctl_timer;

	unsigned int mu_status;
	int out_req_cnt;
	u8 irq;
	unsigned int cardtype;
};

static const char console_inq_page[] =
{
	0x03,0x00,0x03,0x03,0xFA,0x00,0x00,0x30,
	0x50,0x72,0x6F,0x6D,0x69,0x73,0x65,0x20,	/* "Promise " */
	0x52,0x41,0x49,0x44,0x20,0x43,0x6F,0x6E,	/* "RAID Con" */
	0x73,0x6F,0x6C,0x65,0x20,0x20,0x20,0x20,	/* "sole    " */
	0x31,0x2E,0x30,0x30,0x20,0x20,0x20,0x20,	/* "1.00    " */
	0x53,0x58,0x2F,0x52,0x53,0x41,0x46,0x2D,	/* "SX/RSAF-" */
	0x54,0x45,0x31,0x2E,0x30,0x30,0x20,0x20,	/* "TE1.00  " */
	0x0C,0x20,0x20,0x20,0x20,0x20,0x20,0x20
};

static const char mode_sense_cache[] =
{
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4
};

static int gNumAdapter;
static int gReleasedAdapter;
struct st_hba *gShasta_hostp[ST_MAX_ADAPTORS] = {0};

#define MAX_ADAPTER_TYPE					4
static unsigned int   gVenID[MAX_ADAPTER_TYPE] = {
	PCI_VENDOR_ID_PROMISE,
	PCI_VENDOR_ID_PROMISE,
	PCI_VENDOR_ID_PROMISE,
	0x1725
	};
static unsigned int   gDevID[MAX_ADAPTER_TYPE] = {
	0x4302,
	0xc350,
	0x8350,
	0x7250
	};

static int stex_proc_info(char *, char **, off_t, int, int, int);
static int stex_detect(Scsi_Host_Template *);
static int stex_delete(struct Scsi_Host *);
static int stex_shutdown(struct notifier_block *, unsigned long, void *);

static struct notifier_block shasta_notifier = {
	stex_shutdown, NULL, 0
};
MODULE_AUTHOR("Ed Lin");
MODULE_DESCRIPTION("Promise Technology SuperTrak EX Controllers");
MODULE_LICENSE("GPL");


static inline void stex_gettime(u32 *time)
{
	struct timeval tv;
	do_gettimeofday(&tv);

	*time = cpu_to_le32(tv.tv_sec & 0xffffffff);
	*(time + 1) = cpu_to_le32((tv.tv_sec >> 16) >> 16);
}

static inline u16 stex_alloc_tag(unsigned long *bitmap)
{
	int i;
	i = find_first_zero_bit(bitmap, TAG_BITMAP_LENGTH);
	if (i < TAG_BITMAP_LENGTH)
		__set_bit(i, bitmap);
	return (u16)i;
}

static inline void stex_free_tag(unsigned long *bitmap, u16 tag)
{
	clear_bit((int)tag, bitmap);
}

static inline struct status_msg *stex_get_status(struct st_hba *hba)
{
	struct status_msg *status =
		hba->status_buffer + hba->status_tail;

	++hba->status_tail;
	hba->status_tail %= MU_STATUS_COUNT;

	return status;
}

static inline struct req_msg *stex_alloc_req(struct st_hba *hba)
{
	struct req_msg *req = ((struct req_msg *)hba->dma_mem) +
		hba->req_head;

	++hba->req_head;
	hba->req_head %= MU_REQ_COUNT;

	return req;
}

static inline void stex_map_sg(struct st_hba *hba,
	struct req_msg *req, struct scsi_cmnd *cmd)
{
	struct pci_dev *pdev = hba->pdev;
	dma_addr_t dma_handle;
	struct scatterlist *src;
	struct st_sgtable *dst;
	int i;

	dst = (struct st_sgtable *)req->variable;
	dst->max_sg_count = cpu_to_le16(ST_MAX_SG);
	dst->sz_in_byte = cpu_to_le32(cmd->request_bufflen);

	if (cmd->use_sg) {
		src = (struct scatterlist *) cmd->request_buffer;
		dst->sg_count = cpu_to_le16((u16)pci_map_sg(pdev, src,
			cmd->use_sg, cmd->sc_data_direction));

		for (i = 0; i < dst->sg_count; i++, src++) {
			dst->table[i].count = cpu_to_le32((u32)sg_dma_len(src));
			dst->table[i].addr =
				cpu_to_le32(sg_dma_address(src) & 0xffffffff);
			dst->table[i].addr_hi =
				cpu_to_le32((sg_dma_address(src) >> 16) >> 16);
			dst->table[i].ctrl = SG_CF_64B | SG_CF_HOST;
    		}
		dst->table[--i].ctrl |= SG_CF_EOT;
		return;
	}

	dma_handle = pci_map_single(pdev, cmd->request_buffer,
		cmd->request_bufflen, cmd->sc_data_direction);
	cmd->SCp.dma_handle = dma_handle;

	dst->sg_count = cpu_to_le16(1);
	dst->table[0].addr = cpu_to_le32(dma_handle & 0xffffffff);
	dst->table[0].addr_hi = cpu_to_le32((dma_handle >> 16) >> 16);
	dst->table[0].count = cpu_to_le32((u32)cmd->request_bufflen);
	dst->table[0].ctrl = SG_CF_EOT | SG_CF_64B | SG_CF_HOST;
}

static int stex_direct_cp(struct scsi_cmnd *cmd,
	const void *src, unsigned int len)
{
	void *dest;
	unsigned int cp_len;
	struct scatterlist *sg = cmd->request_buffer;

		dest = cmd->request_buffer;
		cp_len = min(cmd->request_bufflen, len);
	memcpy(dest, src, cp_len);
	return cp_len == len;
}

static void stex_controller_info(struct st_hba *hba, struct st_ccb *ccb)
{
	struct st_frame *p;
	struct scsi_cmnd *cmd = ccb->cmd;

		p = cmd->request_buffer;

	memset(p->base, 0, sizeof(u32)*6);
	*(unsigned long *)(p->base) = pci_resource_start(hba->pdev, 0);
	p->rom_addr = 0;

	p->drv_ver.major = ST_VER_MAJOR;
	p->drv_ver.minor = ST_VER_MINOR;
	p->drv_ver.oem = ST_OEM;
	p->drv_ver.build = ST_BUILD_VER;

	p->bus = hba->pdev->bus->number;
	p->slot = hba->pdev->devfn;
	p->irq_level = 0;
	p->irq_vec = hba->pdev->irq;
	p->id = hba->pdev->vendor << 16 | hba->pdev->device;
	p->subid =
		hba->pdev->subsystem_vendor << 16 | hba->pdev->subsystem_device;
}

static void
stex_send_cmd(struct st_hba *hba, struct req_msg *req, u16 tag)
{
	req->tag = cpu_to_le16(tag);
	req->task_attr = TASK_ATTRIBUTE_SIMPLE;
	req->task_manage = 0; /* not supported yet */
	req->payload_sz = (u8)(sizeof(struct req_msg)/sizeof(u32));

	hba->ccb[tag].req = req;
	hba->out_req_cnt++;

	writel(hba->req_head, hba->mmio_base + IMR0);
	writel(MU_INBOUND_DOORBELL_REQHEADCHANGED, hba->mmio_base + IDBL);
	readl(hba->mmio_base + IDBL); /* flush */
}

static int
stex_queuecommand(struct scsi_cmnd *cmd, void (* fn)(struct scsi_cmnd *))
{
	struct st_hba *hba;
	struct Scsi_Host *host;
	unsigned int id,lun;
	struct req_msg *req;
	u16 tag;
	host = cmd->device->host;
	id = cmd->device->id;
	lun = cmd->device->channel; /* firmware lun issue work around */
	hba = (struct st_hba *) &host->hostdata[0];

	switch (cmd->cmnd[0]) {
	case READ_6:
	case WRITE_6:
		cmd->cmnd[9] = 0;
		cmd->cmnd[8] = cmd->cmnd[4];
		cmd->cmnd[7] = 0;
		cmd->cmnd[6] = 0;
		cmd->cmnd[5] = cmd->cmnd[3];
		cmd->cmnd[4] = cmd->cmnd[2];
		cmd->cmnd[3] = cmd->cmnd[1] & 0x1f;
		cmd->cmnd[2] = 0;
		cmd->cmnd[1] &= 0xe0;
		cmd->cmnd[0] += READ_10 - READ_6;
		break;
	case MODE_SENSE:
	{
		static char mode_sense[4] = { 3, 0, 0, 0 };

		if (cmd->cmnd[1] == 0x00 && cmd->cmnd[2] == 0x08)
			stex_direct_cp(cmd, mode_sense_cache,
				sizeof(mode_sense_cache)); /* cache enabled */
		else
			stex_direct_cp(cmd, mode_sense, sizeof(mode_sense));
		cmd->result = DID_OK << 16 | COMMAND_COMPLETE << 8;
		fn(cmd);
		return 0;
	}
	case MODE_SENSE_10:
	{
		static char mode_sense10[8] = { 0, 6, 0, 0, 0, 0, 0, 0 };

		stex_direct_cp(cmd, mode_sense10, sizeof(mode_sense10));
		cmd->result = DID_OK << 16 | COMMAND_COMPLETE << 8;
		fn(cmd);
		return 0;
	}
	case INQUIRY:
		if (id != ST_MAX_ARRAY_SUPPORTED || lun != 0)
			break;
		if((cmd->cmnd[1] & INQUIRY_EVPD) == 0) {
			stex_direct_cp(cmd, console_inq_page,
				sizeof(console_inq_page));
			cmd->result = DID_OK << 16 | COMMAND_COMPLETE << 8;
		} else
			cmd->result = DID_ERROR << 16 | COMMAND_COMPLETE << 8;
		fn(cmd);
		return 0;
	case PASSTHRU_CMD:
		if (cmd->cmnd[1] == PASSTHRU_GET_DRVVER) {
			struct st_drvver ver;
			ver.major = ST_VER_MAJOR;
			ver.minor = ST_VER_MINOR;
			ver.oem = ST_OEM;
			ver.build = ST_BUILD_VER;
			ver.signature[0] = PASSTHRU_SIGNATURE;
			ver.console_id = ST_MAX_ARRAY_SUPPORTED;
			ver.host_no = hba->host->host_no;
			cmd->result = stex_direct_cp(cmd, &ver, sizeof(ver)) ?
				DID_OK << 16 | COMMAND_COMPLETE << 8 :
				DID_ERROR << 16 | COMMAND_COMPLETE << 8;
			fn(cmd);
			return 0;
		}
	default:
		break;
	}

	cmd->scsi_done = fn;

	if (unlikely((tag = stex_alloc_tag((unsigned long *)&hba->tag))
		== TAG_BITMAP_LENGTH))
		return SCSI_MLQUEUE_HOST_BUSY;

	req = stex_alloc_req(hba);
	req->lun = lun;
	req->target = id;

	/* cdb */
	memcpy(req->cdb, cmd->cmnd, STEX_CDB_LENGTH);

	if (cmd->sc_data_direction != PCI_DMA_NONE)
		stex_map_sg(hba, req, cmd);

	hba->ccb[tag].cmd = cmd;
	hba->ccb[tag].sense_bufflen = SCSI_SENSE_BUFFERSIZE;
	hba->ccb[tag].sense_buffer = cmd->sense_buffer;
	if (cmd->use_sg) {
		hba->ccb[tag].page_offset =
			((struct scatterlist *)cmd->request_buffer)->offset;
		hba->ccb[tag].page =
			((struct scatterlist *)cmd->request_buffer)->page;
	}
	hba->ccb[tag].req_type = 0;
	stex_send_cmd(hba, req, tag);
	return 0;
}

static inline void stex_unmap_sg(struct st_hba *hba, struct scsi_cmnd *cmd)
{
	if (cmd->sc_data_direction != PCI_DMA_NONE) {
 		if (cmd->use_sg)
			pci_unmap_sg(hba->pdev, cmd->request_buffer,
				cmd->use_sg, cmd->sc_data_direction);
		else
			pci_unmap_single(hba->pdev, cmd->SCp.dma_handle,
				cmd->request_bufflen, cmd->sc_data_direction);
	}
}

static inline void stex_scsi_done(struct st_ccb *ccb)
{
	struct scsi_cmnd *cmd = ccb->cmd;
	int result;

	if (ccb->srb_status == SRB_STATUS_SUCCESS ||  ccb->srb_status == 0) {
		result = ccb->scsi_status;
		switch (ccb->scsi_status) {
		case SAM_STAT_GOOD:
			result |= DID_OK << 16 | COMMAND_COMPLETE << 8;
			break;
		case SAM_STAT_CHECK_CONDITION:
			result |= DRIVER_SENSE << 24;
			break;
		case SAM_STAT_BUSY:
			result |= DID_BUS_BUSY << 16 | COMMAND_COMPLETE << 8;
			break;
		default:
			result |= DID_ERROR << 16 | COMMAND_COMPLETE << 8;
			break;
		}
	}
	else if (ccb->srb_status & SRB_SEE_SENSE)
		result = DRIVER_SENSE << 24 | SAM_STAT_CHECK_CONDITION;
	else switch (ccb->srb_status) {
		case SRB_STATUS_SELECTION_TIMEOUT:
			result = DID_NO_CONNECT << 16 | COMMAND_COMPLETE << 8;
			break;
		case SRB_STATUS_BUSY:
			result = DID_BUS_BUSY << 16 | COMMAND_COMPLETE << 8;
			break;
		case SRB_STATUS_INVALID_REQUEST:
		case SRB_STATUS_ERROR:
		default:
			result = DID_ERROR << 16 | COMMAND_COMPLETE << 8;
			break;
  	}

   	cmd->result = result;
   	cmd->scsi_done(cmd);
}

static inline void stex_copy_data(struct st_ccb *ccb,
	struct status_msg *resp, unsigned int variable)
{
	void *p;
	if (resp->scsi_status == SAM_STAT_GOOD) {
		if (ccb->cmd == NULL)
			return;
			p = ccb->cmd->request_buffer;
		memcpy(p, resp->variable, min(variable,
			ccb->cmd->request_bufflen));
	} else {
		if (ccb->sense_buffer != NULL)
			memcpy(ccb->sense_buffer, resp->variable,
				min(variable, ccb->sense_bufflen));
	}
}

static void stex_mu_intr(struct st_hba *hba, u32 doorbell)
{
	void __iomem *base = hba->mmio_base;
	struct status_msg *resp;
	struct st_ccb *ccb;
	unsigned int size;
	u16 tag;

	if (!(doorbell & MU_OUTBOUND_DOORBELL_STATUSHEADCHANGED))
		return;

	/* status payloads */
	hba->status_head = readl(base + OMR1);
	if (unlikely(hba->status_head >= MU_STATUS_COUNT)) {
		printk(KERN_WARNING STEX_NAME ": invalid status head\n"
		);
		return;
	}

	if (unlikely(hba->mu_status != MU_STATE_STARTED ||
		hba->out_req_cnt <= 0)) {
		hba->status_tail = hba->status_head;
		goto update_status;
	}

	while (hba->status_tail != hba->status_head) {
		resp = stex_get_status(hba);
		tag = le16_to_cpu(resp->tag);
		if (unlikely(tag >= TAG_BITMAP_LENGTH)) {
			printk(KERN_WARNING STEX_NAME
				": invalid tag\n");
			continue;
		}
		if (unlikely((hba->tag & (1 << tag)) == 0)) {
			printk(KERN_WARNING STEX_NAME
				": null tag\n");
			continue;
		}

		hba->out_req_cnt--;
		ccb = &hba->ccb[tag];
		if (hba->wait_ccb == ccb)
			hba->wait_ccb = NULL;
		if (unlikely(ccb->req == NULL)) {
			printk(KERN_WARNING STEX_NAME
				": lagging req\n");
			stex_free_tag((unsigned long *)&hba->tag, tag);
  			stex_unmap_sg(hba, ccb->cmd); /* ??? */
			continue;
		}

		size = resp->payload_sz * sizeof(u32); /* payload size */
		if (unlikely(size < sizeof(*resp) - STATUS_VAR_LEN ||
			size > sizeof(*resp))) {
			printk(KERN_WARNING STEX_NAME ": bad status size\n");
		} else {
			size -= sizeof(*resp) - STATUS_VAR_LEN; /* copy size */
			if (size)
				stex_copy_data(ccb, resp, size);
		}

		ccb->srb_status = resp->srb_status;
		ccb->scsi_status = resp->scsi_status;

		if (ccb->req_type & PASSTHRU_REQ_TYPE) {
			if (ccb->req_type & PASSTHRU_REQ_NO_WAKEUP) {
				ccb->req_type = 0;
				continue;
			}
			ccb->req_type = 0;
			if (waitqueue_active(&hba->waitq))
				wake_up(&hba->waitq);
			continue;
		}
		if (ccb->cmd->cmnd[0] == PASSTHRU_CMD &&
			ccb->cmd->cmnd[1] == PASSTHRU_GET_ADAPTER)
			stex_controller_info(hba, ccb);
		stex_free_tag((unsigned long *)&hba->tag, tag);
  		stex_unmap_sg(hba, ccb->cmd);
		stex_scsi_done(ccb);
	}

update_status:
	writel(hba->status_head, base + IMR1);
	readl(base + IMR1); /* flush */
}

static inline void  stex_intr(int irq, void *__hba, struct pt_regs *regs)
{
	struct st_hba *hba = __hba;
	void __iomem *base = hba->mmio_base;
	u32 data;
	unsigned long flags;
	int handled = 0;

	spin_lock_irqsave(&io_request_lock, flags);

	data = readl(base + ODBL);

	if (data && data != 0xffffffff) {
		/* clear the interrupt */
		writel(data, base + ODBL);
		readl(base + ODBL); /* flush */
		stex_mu_intr(hba, data);
		handled = 1;
	}

	spin_unlock_irqrestore(&io_request_lock, flags);


}

static void stex_timeout(wait_queue_head_t *waitq_p)
{
   	if (waitqueue_active(waitq_p))
		wake_up(waitq_p);
}

static void stex_add_timer(struct st_hba *hba, unsigned long data, int sched)
{
   	hba->ioctl_timer.data = data;
   	hba->ioctl_timer.function = (void (*)(unsigned long))stex_timeout;
	mod_timer(&hba->ioctl_timer, jiffies + sched);
}

static int stex_handshake(struct st_hba *hba)
{
	void __iomem *base = hba->mmio_base;
	struct handshake_frame *h;
	dma_addr_t status_phys;
	int i;

	if (readl(base + OMR0) != MU_HANDSHAKE_SIGNATURE) {
		writel(MU_INBOUND_DOORBELL_HANDSHAKE, base + IDBL);
		readl(base + IDBL);
		for (i = 0; readl(base + OMR0) != MU_HANDSHAKE_SIGNATURE
			&& i < MU_MAX_DELAY_TIME; i++) {
			rmb();
			mdelay(1);
		}

		if (i == MU_MAX_DELAY_TIME) {
			printk(KERN_ERR STEX_NAME
				": no handshake signature\n");
			return -1;
		}
	}

	udelay(10);

	h = (struct handshake_frame *)(hba->dma_mem + MU_REQ_BUFFER_SIZE);
	h->rb_phy = cpu_to_le32(hba->dma_handle);
	h->rb_phy_hi = cpu_to_le32((hba->dma_handle >> 16) >> 16);
	h->req_sz = cpu_to_le16(sizeof(struct req_msg));
	h->req_cnt = cpu_to_le16(MU_REQ_COUNT);
	h->status_sz = cpu_to_le16(sizeof(struct status_msg));
	h->status_cnt = cpu_to_le16(MU_STATUS_COUNT);
	stex_gettime(&h->hosttime);
	h->partner_type = HMU_PARTNER_TYPE;

	status_phys = hba->dma_handle + MU_REQ_BUFFER_SIZE;
	writel(status_phys, base + IMR0);
	readl(base + IMR0);
	writel((status_phys >> 16) >> 16, base + IMR1);
	readl(base + IMR1);

	writel((status_phys >> 16) >> 16, base + OMR0); /* old fw compatible */
	readl(base + OMR0);
	writel(MU_INBOUND_DOORBELL_HANDSHAKE, base + IDBL);
	readl(base + IDBL); /* flush */

	udelay(10);
	for (i = 0; readl(base + OMR0) != MU_HANDSHAKE_SIGNATURE
		&& i < MU_MAX_DELAY_TIME; i++) {
		rmb();
		mdelay(1);
	}

	if (i == MU_MAX_DELAY_TIME) {
		printk(KERN_ERR STEX_NAME
			": no signature after handshake frame\n");
		return -1;
	}

	writel(0, base + IMR0);
	readl(base + IMR0);
	writel(0, base + OMR0);
	readl(base + OMR0);
	writel(0, base + IMR1);
	readl(base + IMR1);
	writel(0, base + OMR1);
	readl(base + OMR1); /* flush */
	hba->mu_status = MU_STATE_STARTED;
	return 0;
}

static int stex_abort(struct scsi_cmnd *cmd)
{
	struct Scsi_Host *host = cmd->device->host;
	struct st_hba *hba = (struct st_hba *)host->hostdata;
	u16 tag;
	void __iomem *base = hba->mmio_base;
	u32 data;
	int result = SUCCESS;
	for (tag = 0; tag < MU_MAX_REQUEST; tag++)
		if (hba->ccb[tag].cmd == cmd && (hba->tag & (1 << tag))) {
			hba->wait_ccb = &(hba->ccb[tag]);
			break;
		}
	if (tag >= MU_MAX_REQUEST)
		goto out;

	data = readl(base + ODBL);
	if (data == 0 || data == 0xffffffff)
		goto fail_out;

	writel(data, base + ODBL);
	readl(base + ODBL); /* flush */

	stex_mu_intr(hba, data);

 	if (hba->wait_ccb == NULL) {
 		printk(KERN_WARNING STEX_NAME
			": lost interrupt\n");
		goto out;
	}

fail_out:
	hba->wait_ccb->req = NULL; /* nullify the req's future return */
	hba->wait_ccb = NULL;
	result = FAILED;
out:
	return result;
}

static void stex_hard_reset(struct st_hba *hba)
{

	struct pci_bus *bus;
	int i;
	u16 pci_cmd;
	u8 pci_bctl;
        u32	saved_config_space[16];
	for (i = 0; i < 16; i++)
		pci_read_config_dword(hba->pdev, i * 4,
			&saved_config_space[i]);

	/* Reset secondary bus. Our controller(MU/ATU) is the only device on
	   secondary bus. Consult Intel 80331/3 developer's manual for detail */
	bus = hba->pdev->bus;
	pci_read_config_byte(bus->self, PCI_BRIDGE_CONTROL, &pci_bctl);
	pci_bctl |= PCI_BRIDGE_CTL_BUS_RESET;
	pci_write_config_byte(bus->self, PCI_BRIDGE_CONTROL, pci_bctl);
	msleep(1);
	pci_bctl &= ~PCI_BRIDGE_CTL_BUS_RESET;
	pci_write_config_byte(bus->self, PCI_BRIDGE_CONTROL, pci_bctl);

	for (i = 0; i < MU_MAX_DELAY_TIME; i++) {
		pci_read_config_word(hba->pdev, PCI_COMMAND, &pci_cmd);
		if (pci_cmd & PCI_COMMAND_MASTER)
			break;
		msleep(1);
	}
	msleep(5000);
	for (i = 0; i < 16; i++)
		pci_write_config_dword(hba->pdev, i * 4,
			saved_config_space[i]);

	writel(MU_INBOUND_DOORBELL_HANDSHAKE, hba->mmio_base + IDBL);
	readl(hba->mmio_base + IDBL);
       msleep(3000);

}

static int stex_reset(struct scsi_cmnd *cmd)
{
	struct st_hba *hba;
	int tag;
	int i = 0;
	hba = (struct st_hba *) &cmd->device->host->hostdata[0];

wait_cmds:
	for (tag = 0; tag < MU_MAX_REQUEST; tag++)
		if ((hba->tag & (1 << tag)) && hba->ccb[tag].req != NULL)
			break;
	spin_unlock_irq(&io_request_lock);
	if (tag < MU_MAX_REQUEST) {
		mdelay(1000);
		if (++i < 10) {
	spin_lock_irq(&io_request_lock);
			goto wait_cmds;
		}
	}

	hba->mu_status = MU_STATE_RESETTING;
	writel(MU_INBOUND_DOORBELL_HANDSHAKE, hba->mmio_base + IDBL);
	readl(hba->mmio_base + IDBL);
	mdelay(3000);

	spin_lock_irq(&io_request_lock);

	for (tag = 0; tag < MU_MAX_REQUEST; tag++)
		if ((hba->tag & (1 << tag)) && hba->ccb[tag].req != NULL) {
			stex_free_tag((unsigned long *)&hba->tag, tag);
			stex_unmap_sg(hba, hba->ccb[tag].cmd);
			hba->ccb[tag].cmd->result = DID_RESET << 16;
			hba->ccb[tag].cmd->scsi_done(hba->ccb[tag].cmd);
		}

	spin_unlock_irq(&io_request_lock);


	if (hba->cardtype == st_shasta)
		stex_hard_reset(hba);
	if (stex_handshake(hba)) {
		spin_lock_irq(&io_request_lock);
		printk(KERN_WARNING STEX_NAME
			": resetting: handshake failed\n");
		return FAILED;
	}
	spin_lock_irq(&io_request_lock);
	hba->tag = 0;
	hba->req_head = 0;
	hba->req_tail = 0;
	hba->status_head = 0;
	hba->status_tail = 0;
	hba->out_req_cnt = 0;
	return SUCCESS;
}

static void stex_init_hba(struct st_hba *hba,
	struct Scsi_Host * host, struct pci_dev *pdev)
{
	host->max_channel = ST_MAX_LUN_PER_TARGET; /* fw lun work around */
    	host->max_id = ST_MAX_TARGET_NUM;
    	host->max_lun = 1; /* fw lun work around */
	host->unique_id = host->host_no;
	host->max_cmd_len = STEX_CDB_LENGTH;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,4) && \
	LINUX_VERSION_CODE  < KERNEL_VERSION(2,5,0)
	scsi_set_pci_device(host, pdev);
#endif
	hba->host = host;
	hba->pdev = pdev;
	init_waitqueue_head(&hba->waitq);
	init_timer(&hba->ioctl_timer);
}

static int stex_init_shm(struct st_hba *hba, struct pci_dev *pdev)
{
	hba->dma_mem = pci_alloc_consistent(pdev,
		MU_BUFFER_SIZE, &hba->dma_handle);
	if (!hba->dma_mem)
		return -ENOMEM;

	hba->status_buffer =
		(struct status_msg *)(hba->dma_mem + MU_REQ_BUFFER_SIZE);
	hba->mu_status = MU_STATE_STARTING;
	return 0;
}


static int stex_biosparam(Disk *disk, kdev_t dev, int geom[])
{
	unsigned  long capacity = disk->capacity;
	int heads = 255, sectors = 63, cylinders;

	if (capacity < 0x200000) {
		heads = 64;
		sectors = 32;
	}
	cylinders = capacity/(heads * sectors);
	geom[0] = heads;
	geom[1] = sectors;
	geom[2] = cylinders;

	return 0;
}


static void stex_internal_flush(struct st_hba *hba, int id, u16 tag)
{
	struct req_msg *req;

	req = stex_alloc_req(hba);
	memset(req->cdb, 0, STEX_CDB_LENGTH);

	if (id < ST_MAX_ARRAY_SUPPORTED*ST_MAX_LUN_PER_TARGET) {
		req->target = id/ST_MAX_LUN_PER_TARGET;
		req->lun = id%ST_MAX_LUN_PER_TARGET;
		req->cdb[0] = CONTROLLER_CMD;
		req->cdb[1] = CTLR_POWER_STATE_CHANGE;
		req->cdb[2] = CTLR_POWER_SAVING;
	} else {
		req->target = id/ST_MAX_LUN_PER_TARGET - ST_MAX_ARRAY_SUPPORTED;
		req->lun = id%ST_MAX_LUN_PER_TARGET;
   		req->cdb[0] = SYNCHRONIZE_CACHE;
	}

	hba->ccb[tag].cmd = NULL;
	hba->ccb[tag].page_offset = 0;
 	hba->ccb[tag].page = NULL;
	hba->ccb[tag].sense_bufflen = 0;
	hba->ccb[tag].sense_buffer = NULL;
	hba->ccb[tag].req_type |= PASSTHRU_REQ_TYPE;

	stex_send_cmd(hba, req, tag);
}
	
#if defined(_SUPPORT_DISKDUMP) || defined(CONFIG_DISKDUMP) || defined(CONFIG_DISKDUMP_MODULE)
static int stex_sanity_check(struct scsi_device *sdev)
{
	struct st_hba *hba;

	if (!sdev->host || !sdev->host->hostdata)
		return -ENXIO;

	hba = (struct st_hba *)sdev->host->hostdata;

	if (hba->mu_status != MU_STATE_STARTED)
		return -EIO;

	return 0;
}

static void stex_poll(struct scsi_device *sdev)
{
	struct st_hba *hba;
	void __iomem *base;
	u32 data;

	hba = (struct st_hba *)sdev->host->hostdata;

	spin_lock(&io_request_lock);
	
	base = hba->mmio_base;

	data = readl(base + ODBL);
	
	if (data && data != 0xffffffff) {
		writel(data, base + ODBL);
		readl(base + ODBL);
		stex_mu_intr(hba, data);
	}

	spin_unlock(&io_request_lock);
}


#endif //DISKDUMP


static int stex_set_dma_mask(struct pci_dev * pdev, int *p_hiio)
{
	*p_hiio = 0;
#if defined(_64BPLATFORM) || defined(CONFIG_HIGHMEM64G)
	if (pci_set_dma_mask(pdev, 0xffffffffffffffffULL) == 0) {
		*p_hiio = 1;
		return 0;
	}
#endif
	return(pci_set_dma_mask(pdev, 0xffffffffULL));
}

static void stex_hba_stop(struct st_hba *hba)
{
	unsigned long flags;
	int i;
	u16 tag;

	spin_lock_irqsave(&io_request_lock, flags);
	if ((tag = stex_alloc_tag((unsigned long *)&hba->tag))
		== TAG_BITMAP_LENGTH) {
		spin_unlock_irqrestore(&io_request_lock, flags);
		printk(KERN_ERR STEX_NAME ": unable to alloc tag\n");
		return;
	}
	for (i=0; i<(ST_MAX_ARRAY_SUPPORTED*ST_MAX_LUN_PER_TARGET*2); i++) {
		stex_internal_flush(hba, i, tag);
		spin_unlock_irqrestore(&io_request_lock, flags);

		stex_add_timer(hba, (unsigned long)&hba->waitq,
			ST_INTERNAL_TIMEOUT*HZ);
		wait_event(hba->waitq,!(hba->ccb[tag].req_type));
		del_timer_sync(&hba->ioctl_timer);
		if (hba->ccb[tag].req_type & PASSTHRU_REQ_TYPE)
			return;
		spin_lock_irqsave(&io_request_lock, flags);
	}

	stex_free_tag((unsigned long *)&hba->tag, tag);
	spin_unlock_irqrestore(&io_request_lock, flags);
}

static void stex_hba_free(struct st_hba *hba)
{
	free_irq(hba->pdev->irq, hba);

	iounmap(hba->mmio_base);

	pci_release_regions(hba->pdev);

	if (hba->dma_mem)
		pci_free_consistent(hba->pdev, MU_BUFFER_SIZE,
			hba->dma_mem, hba->dma_handle);
}



static int stex_proc_info(char *buffer, char **start, off_t offset,
			int length, int hostno, int func)
{
	int len;

	*start = buffer;
	if (func == TRUE) {
		return -EINVAL;
	}

	len = sprintf(buffer,
		"***** SuperTrak EX8350/8300/16350/16300 Driver %s %s*****\n",
		ST_DRIVER_VERSION, ST_DRIVER_DATE);

	len += sprintf(buffer + len,
		"***** Copyright 2005 by Promise Technology, Inc.   *****\n\n");

	*(buffer + len) = '\0';
 
	*start = buffer + offset;
	len -= offset;
	if(len > length) {
		len = length;
	} else if(len < 0){
		len = 0;
		**start = '\0';
	}
	return len;
}

static void stex_ha_init(struct pci_dev *pdev, Scsi_Host_Template *template,unsigned int  cardtype)
{
	int err, himem_io;
	struct st_hba *hba;
	struct Scsi_Host * host;

	if(gNumAdapter >= ST_MAX_ADAPTORS){
		printk(KERN_WARNING "too many cards present\n");
		return;
	}

	err = pci_enable_device(pdev);
	if (err) return;

	pci_set_master(pdev);
	
	host = scsi_register(template, sizeof(struct st_hba));
	
	if (!host) {
		printk(KERN_ERR STEX_NAME ": scsi_register failed\n");
		err = -ENOMEM;
		goto out_disable;
	}
	
	hba = (struct st_hba *)host->hostdata;
	memset(hba, 0, sizeof(struct st_hba));

	err = pci_request_regions(pdev, STEX_NAME);
	if (err < 0) {
		printk(KERN_ERR STEX_NAME ": request regions failed\n");
		goto out_scsi_host_put;
	}

	hba->mmio_base = ioremap(pci_resource_start(pdev, 0), 
				pci_resource_len(pdev, 0));
	if ( !hba->mmio_base) {
		printk(KERN_ERR STEX_NAME ": memory map failed\n");
		err = -ENOMEM;
		goto out_release_regions;
	}

	err = stex_set_dma_mask(pdev, &himem_io);
	if (err) {
		printk(KERN_ERR STEX_NAME ": set dma mask failed\n");
		goto out_iounmap;
	}
	if (himem_io == 0)
		template->highmem_io = 0;

	err = stex_init_shm(hba, pdev);
	if (err) {
		printk(KERN_ERR STEX_NAME ": dma mem alloc failed\n");
		goto out_iounmap;
	}


	hba->cardtype = cardtype;
	stex_init_hba(hba, host, pdev);	

	err = request_irq(pdev->irq, stex_intr, SA_SHIRQ, STEX_NAME, hba);
	if (err) {
		printk(KERN_ERR STEX_NAME ": request irq failed\n");
		goto out_pci_free;
	}
	pci_set_drvdata(pdev, hba);
	gShasta_hostp[gNumAdapter] = hba;
	gNumAdapter++;	

	return;

out_pci_free:
	pci_free_consistent(pdev, MU_BUFFER_SIZE,
		hba->dma_mem, hba->dma_handle);
out_iounmap:
	iounmap(hba->mmio_base);	
out_release_regions:
	pci_release_regions(pdev);
out_scsi_host_put:
	scsi_unregister(host);
out_disable:
	pci_disable_device(pdev);	
}

static int stex_detect(Scsi_Host_Template * SHT)
{
	struct pci_dev *pdev = NULL;
	struct st_hba *hba;
	int cnt, cnt1;
	int err;
	unsigned int cardtype;

	SHT->highmem_io = 1;
	for (cnt = 0; cnt < MAX_ADAPTER_TYPE; cnt++)
		while ((pdev=pci_find_device(gVenID[cnt], gDevID[cnt], pdev))) {
		  if(gVenID[cnt]==0x1725&&gVenID[cnt]==0x7250)
                  cardtype=st_vsc;
	           else
                  cardtype=st_shasta; 
		stex_ha_init(pdev, SHT,cardtype);
		}

	for (cnt = 0; cnt < gNumAdapter; cnt++) {
		hba = gShasta_hostp[cnt];
		err = stex_handshake(hba);
		if (err) {
			pdev = hba->pdev;
			free_irq(hba->irq, (void *)hba);
			iounmap(hba->mmio_base);	
			pci_release_regions(pdev);
			scsi_unregister(hba->host);
			pci_disable_device(pdev);
			gShasta_hostp[cnt] = NULL;
			gReleasedAdapter ++;
	}

		register_reboot_notifier(&shasta_notifier);

	}

	return (gNumAdapter - gReleasedAdapter);
}

static int stex_delete(struct Scsi_Host *host)
{
	struct st_hba *hba;
	int i;

	hba = (struct st_hba *)host->hostdata;
	if (!hba) {
		printk(KERN_CRIT STEX_NAME ": invalid detach\n");
		return 0;
	}

	for (i = 0; i < gNumAdapter && gShasta_hostp[i] != hba; i++) ;

	if (i == gNumAdapter) {
		printk(KERN_WARNING STEX_NAME
		       ":release, invalid hostdata\n");
		return 0;
	}

	pci_set_drvdata(hba->pdev, NULL);

	stex_hba_stop(hba);

	stex_hba_free(hba);

	scsi_unregister(hba->host);

	gShasta_hostp[i] = NULL;
	if (gReleasedAdapter == gNumAdapter) 
		unregister_reboot_notifier(&shasta_notifier);
	return 0;
}

static int
stex_shutdown(struct notifier_block *nb, unsigned long event, void *buf)
{
	struct st_hba *hba;
	int i;

	if ((event != SYS_RESTART) && (event != SYS_HALT) &&
	    (event != SYS_POWER_OFF))
		return (NOTIFY_DONE);

	for (i = 0; i < gNumAdapter; i++) {
		hba = gShasta_hostp[i];

		if (!hba) continue;

		stex_hba_stop(hba);
	}

	return (NOTIFY_OK);
}
#define _STEX_SHT	\
	.module			= THIS_MODULE,				\
	.name			= "Promise SuperTrak EX driver",	\
	.proc_name		= STEX_NAME,				\
	.bios_param		= stex_biosparam,			\
	.queuecommand		= stex_queuecommand,			\
	.eh_abort_handler	= stex_abort,				\
	.eh_host_reset_handler	= stex_reset,				\
	.can_queue		= ST_CAN_QUEUE - 1,			\
	.this_id		= -1,					\
	.sg_tablesize		= ST_MAX_SG,				\
	.cmd_per_lun		= ST_CMD_PER_LUN, 	\
	.detect			= stex_detect,			\
	.release		= stex_delete,			\
	.proc_info		= stex_proc_info,			\
	.use_new_eh_code	= 1					\

#if defined(CONFIG_DISKDUMP) || defined(CONFIG_DISKDUMP_MODULE)

static struct scsi_dump_ops stex_dump_ops = {
	.sanity_check		= stex_sanity_check,
	.poll			= stex_poll,
	.no_write_cache_enable	= 1,
};
static Scsi_Host_Template_dump driver_template_dump = {
	.hostt = {
		_STEX_SHT,
		.disk_dump		= 1
	},
	.dump_ops			= &stex_dump_ops
};
#define driver_template	(driver_template_dump.hostt)
#else

static Scsi_Host_Template driver_template = { _STEX_SHT };
#endif

#include "scsi_module.c"

